#include <iostream>
#include <cassert>
#include "graph.hpp"
#include "trees.hpp"

void majorTests();
bool minorTests();
bool validityTests();

// Adapted from GeeksforGeeks, thanks to the orig authors. :)  
void CombinationRepetitionUtil(std::vector<int> &chosen, std::vector<std::vector<int>> &combos, int index, int options, int start, int end) {
    if (index == options) {
        std::vector<int> combo(options, 0);
        for (int i = 0; i < options; i++)
            combo[i] = chosen[i];
        combos.push_back(combo);
        return;
    }
  
    for (int i = start; i <= end; i++) {
        chosen[index] = i;
        CombinationRepetitionUtil(chosen, combos, index + 1, options, i, end);
    }
}
  
void nonRedundantCombos(std::vector<std::vector<int>> &combos, int options, int length) {
    std::vector<int> chosen(length+1, 0);
    CombinationRepetitionUtil(chosen, combos, 0, length, 0, options);
}

bool forestBuilder(Graph &forest, Graph &inputGraph, std::vector<int> const &treeList, int index, const int treeNum, const bool stopAtReduction) {
    if (index >= treeNum) {
        std::vector<bool> finalists;
        //do {
            // std::cout<<"Trying ";
            // forest.printGraph();
            finalists = Graph::findFactor(inputGraph, forest);

            if (!finalists.empty()) {
                std::cout<<"This reduction: " <<std::endl;
                forest.printGraph(); 
                std::cout<<"Reduces: ";
                inputGraph.printGraph();

                // This boolean decides whether we look for all reductions or stop at first.
                return stopAtReduction;
            }
        //} while (forest.permuteEdges(treeList));

        return false;
    }
    
    std::vector<std::vector<int>> combos;
    int similar = index;
    while (similar < treeNum && (treeList[similar] == treeList[similar+1])) { similar++; };

    nonRedundantCombos(combos,forest::treeGraphs[treeList[index] - 2].size()-1, similar - index + 1);
    index = similar;

    for (int i = 0; i < combos.size(); i++) {
        // std::cout<<"Testing combo number " <<i <<" of " <<combos.size()-1 <<std::endl;
        
        Graph grownForest(forest);
        for (int j = 0; j <combos[i].size(); j++) {
            grownForest.expandGraph(forest::treeGraphs[treeList[index] - 2][combos[i][j]], forest::treeMasks[treeList[index] - 2][combos[i][j]]);
        }

        //std::cout<<std::endl; grownForest.printGraph(); std::cout<<std::endl;           
        if (forestBuilder(grownForest, inputGraph, treeList, index+1, treeNum, stopAtReduction)) return true;
    } 

    return false;
}

bool treeTest(Graph &inputGraph, bool stopAtReduction) {
    //inputGraph.outVertCount is the number I'm working with,
    //find all possible ways of creating it, cover them with -2 indexes into treeGraphs/Masks
    std::vector<std::vector<std::vector<int>>> combinations = {
        {{2}},{{3}},{{2,2},{4}},{{2,3},{5}},{{2,2,2},{2,4},{3,3},{6}},{{2,2,3},{2,5},{3,4},{7}},{{2,2,2,2},{2,2,4},{2,3,3},{2,6},{3,5},{4,4}}, // Excluded Extras {8}
        {{2,2,2,3},{2,3,4},{3,3,3},{2,2,5},{2,7},{3,6},{4,5}}, // Excluded  Extras {9}
        {{2,2,2,2,2},{2,2,2,4},{2,2,3,3},{2,2,6},{2,3,5},{2,4,4},{3,3,4},{3,7},{4,6},{5,5}},// Extras {10},{2,8}
        {{2,2,2,2,3},{2,2,2,5},{2,2,3,4},{2,2,7},{2,3,3,3},{2,3,6},{2,4,5},{3,3,5},{3,4,4},{4,7},{5,6}}, // Extras {11},{3,8},{2,9}
        {{2,2,2,2,2,2},{2,2,2,2,4},{2,2,2,3,3},{2,2,2,6},{2,2,3,5},{2,2,4,4},{2,3,3,4},{2,3,7},{2,4,6},{2,5,5},{3,3,3,3},{3,3,6},{3,4,5},{4,4,4},{5,7},{6,6}}, // Extras {12},{4,8},{3,9},{2,10},{2,2,8}
        {{2,2,2,2,2,3},{2,2,2,2,5},{2,2,2,3,4},{2,2,2,7},{2,2,3,3,3},{2,2,3,6},{2,2,4,5},{2,3,3,5},{2,3,4,4},{2,4,7},{2,5,6},{3,3,3,4},{3,3,7},{3,4,6},{3,5,5},{4,4,5},{6,7}} // Extras {2,2,9} {2,3,8} {2,11} {3,10} {4,9} {5,8} {13}
    };
    for (auto row : combinations[inputGraph.getOutVertCount() - 2]) {
        Graph starter(0,0,{},{});
        if (forestBuilder(starter, inputGraph, row, 0, row.size(), stopAtReduction)) return true;
        // std::cout<<"We advance to the next row of combinations" <<std::endl;
    }

    return false;
}

void groupTest(std::vector<std::vector<int>> const &inputGraphs, std::vector<std::vector<int>> const &inputMasks, const bool stopAtReduction) {
    for (int i = 0; i < inputGraphs.size(); i++) {
        Graph cyclus(inputMasks[i][0],inputMasks[i][1], inputGraphs[i], {});

        std::cout<<std::endl <<"Testing this: "; cyclus.printGraph();
        treeTest(cyclus, stopAtReduction);
    }
}

int main() {
    bool works = true;

    works = validityTests();
    assert(works);

    works = minorTests();
    assert(works);

    majorTests();

    return 0;
}

void majorTests() {

    std::vector<std::vector<int>> tenCycle = {
        {19,-1,-1,  10,-1,-1,  11,-1,-1,  12,-1,-1,  13,-1,-1,  14,-1,-1,  15,-1,-1,  16,-1,-1,  17,-1,-1,  18,-1,-1,  1,11,19,  2,10,12,  3,11,13,  4,12,14,  5,13,15,  6,14,16,  7,15,17,  8,16,18,  9,17,19,  0,10,18,  },
    };
    std::vector<std::vector<int>> tenCycleOuts = {
        {20,10},
    };

    // std::cout<<"Testing 10-10 cycle:" <<std::endl;
    // groupTest(tenCycle, tenCycleOuts, true);

    std::vector<std::vector<int>> altSevenCycle_threeMixed = {
        {17,-1,-1,  17,-1,-1,  18,-1,-1,  18,-1,-1,  19,-1,-1,  19,-1,-1,  13,-1,-1,  14,-1,-1,  15,-1,-1,  16,-1,-1,  11,16,17,  10,12,18,  11,13,19,  6,12,14,  7,13,15,  8,14,16,  9,10,15,  0,1,10,  2,3,11,  4,5,12,  },
        {17,-1,-1,  17,-1,-1,  18,-1,-1,  18,-1,-1,  19,-1,-1,  19,-1,-1,  12,-1,-1,  14,-1,-1,  15,-1,-1,  16,-1,-1,  11,16,17,  10,12,18,  6,11,13,  12,14,19,  7,13,15,  8,14,16,  9,10,15,  0,1,10,  2,3,11,  4,5,13,  },
        {17,-1,-1,  17,-1,-1,  18,-1,-1,  18,-1,-1,  19,-1,-1,  19,-1,-1,  12,-1,-1,  14,-1,-1,  15,-1,-1,  10,-1,-1,  9,11,16,  10,12,18,  6,11,13,  12,14,19,  7,13,15,  8,14,16,  10,15,17,  0,1,16,  2,3,11,  4,5,13,  }, 
    };
    std::vector<std::vector<int>> altSevenCycleOuts_threeMixed = {
        {20,10},{20,10},{20,10},
    };

    // std::cout<<"Testing triple-extended 7 cycles:" <<std::endl;
    // groupTest(altSevenCycle_threeMixed, altSevenCycleOuts_threeMixed, true);

    std::vector<std::vector<int>> altSevenCycle_twoMixed = {
        {9,-1,-1,  9,-1,-1,  10,-1,-1,  10,-1,-1,  12,-1,-1,  13,-1,-1,  14,-1,-1,  15,-1,-1,  16,-1,-1,  0,1,17,  2,3,11,  10,12,17,  4,11,13,  5,12,14,  6,13,15,  7,14,16,  8,15,17,  9,11,16},
        {9,-1,-1,  9,-1,-1,  10,-1,-1,  10,-1,-1,  11,-1,-1,  13,-1,-1,  14,-1,-1,  15,-1,-1,  16,-1,-1,  0,1,17,  2,3,12,  4,12,17,  10,11,13,  5,12,14,  6,13,15,  7,14,16,  8,15,17,  9,11,16},
    };
    std::vector<std::vector<int>> altSevenCycleOuts_twoMixed = {
        {18,9}, {18,9},
    };

    // std::cout<<"Testing doubly-extended 7 cycles:" <<std::endl;
    // groupTest(altSevenCycle_twoMixed, altSevenCycleOuts_twoMixed, true);

    std::vector<std::vector<int>> altSevenCycle_oneMixed = {
        {11,-1,-1,  10,-1,-1,  9,-1,-1,  15,-1,-1,  14,-1,-1,  13,-1,-1,  8,-1,-1,  8,-1,-1,  6,7,12,  2,10,15,  1,9,11,  0,10,12,  8,11,13,  5,12,14,  4,13,15,  3,9,14,  },
        //With a longer arm
        {9,-1,-1,  9,-1,-1,  11,-1,-1,  12,-1,-1,  13,-1,-1,  14,-1,-1,  15,-1,-1,  16,-1,-1,  10,11,16,  0,1,10,  8,9,-1,  2,8,12,  3,11,13,  4,12,14,  5,13,15,  6,14,16,  7,8,15},
    };
    std::vector<std::vector<int>> altSevenCycleOuts_oneMixed = {
        {16,8},
        {17,8},
    };

    // std::cout<<"Testing extended 7 cycles:" <<std::endl;
    // groupTest(altSevenCycle_oneMixed, altSevenCycleOuts_oneMixed, true);

    std::vector<std::vector<int>> altEightCycle = {
        {9,-1,-1,  9,-1,-1,  11,-1,-1,  12,-1,-1,  13,-1,-1,  14,-1,-1,  15,-1,-1,  16,-1,-1,  17,-1,-1,  0,1,10,  9,11,17,  2,10,12,  3,11,13,  4,12,14,  5,13,15,  6,14,16,  7,15,17,  8,10,16},
    };
    std::vector<std::vector<int>> altEightCycleOuts = {
        {18,9},
    };

    // std::cout<<"Testing the expanded eight-cycle:" <<std::endl;
    // groupTest(altEightCycle, altEightCycleOuts, true);

    std::vector<std::vector<int>> eightCycle = {
        {8,-1,-1,  15,-1,-1,  14,-1,-1,  13,-1,-1,  12,-1,-1,  11,-1,-1,  10,-1,-1,  9,-1,-1,  0,9,15,  7,8,10,  6,9,11,  5,10,12,  4,11,13,  3,12,14,  2,13,15,  1,8,14},  
    };
    std::vector<std::vector<int>> eightCycleOuts = {
        {16,8},
    };

    // std::cout<<"Testing the 8-8 cycle:" <<std::endl;
    // groupTest(eightCycle, eightCycleOuts, false);

    std::vector<std::vector<int>> altSixCycle = {
        {12,-1,-1,  7,-1,-1,  8,-1,-1,  9,-1,-1,  10,-1,-1,  13,-1,-1,  13,-1,-1,  1,8,12,  2,7,9,  3,8,10,  4,9,11,  10,12,13,  0,7,11,  5,6,11},

        {9,-1,-1,  10,-1,-1,  11,-1,-1,  6,-1,-1,  8,-1,-1,  7,-1,-1,  3,8,11,  5,8,9,  4,6,7,  0,7,10,  1,9,11,  2,6,10},
    };
    std::vector<std::vector<int>> altSixCycleOuts = {
        {14,7},
        {12,6}
    };

    // std::cout<<"Testing the extended 6 cycle and original 6-6:" <<std::endl;
    // groupTest(altSixCycle, altSixCycleOuts, false);

    std::vector<std::vector<int>> sevenCycleTest = {
        //Full seven-cycle
        // {7,-1,-1,  13,-1,-1,  8,-1,-1,  9,-1,-1,  10,-1,-1,  11,-1,-1,  12,-1,-1,  0,8,13,  2,7,9,  3,8,10,  4,9,11,  5,10,12,  6,11,13,  1,7,12},
        
        //Seven-cycles with deg 2 edges
        {7,-1,-1,  6,-1,-1,  11,-1,-1,  10,-1,-1,  9,-1,-1,  8,-1,-1,  1,7,12,  0,6,8,  5,7,9,  4,8,10,  3,9,11,  2,10,12,  6,11,-1},
        // {8,-1,-1,  7,-1,-1,  9,-1,-1,  10,-1,-1,  11,-1,-1,  6,11,-1,  5,7,-1,  1,6,8,  0,7,9,  2,8,10,  3,9,11,  4,5,10},
        // {8,-1,-1,  7,-1,-1,  9,-1,-1,  10,-1,-1,  5,10,-1,  4,6,-1,  5,7,-1,  1,6,8,  0,7,9,  2,8,10,  3,4,9},
        // {8,-1,-1,  7,-1,-1,  9,-1,-1,  4,9,-1,  3,5,-1,  4,6,-1,  5,7,-1,  1,6,8,  0,7,9,  2,3,8},
        // {8,-1,-1,  6,-1,-1,  3,-1,-1,  2,4,9,  3,5,-1,  4,6,-1,  1,5,7,  6,8,-1,  0,7,9,  3,8,-1},
        // {8,-1,-1,  6,-1,-1,  3,8,-1,  2,4,-1,  3,5,-1,  4,6,-1,  1,5,7,  6,8,-1,  0,2,7},
    };
    std::vector<std::vector<int>> sevenCycleTestOuts = {
        //{14,7},
        {13,6},//{12,5},{11,4},{10,3},{10,3},{9,2},
    };

    // std::cout<<"Testing the 7-cycles:" <<std::endl;
    // groupTest(sevenCycleTest, sevenCycleTestOuts, false);

    std::vector<std::vector<int>> mixedCycleTest_one = {
        //Full 6-7 cycle with 1 common edge
        {12,-1,-1,  11,-1,-1,  10,-1,-1,  9,-1,-1,  15,-1,-1,  16,-1,-1,  17,-1,-1,  18,-1,-1,  19,-1,-1,  3,10,14,  2,9,11,  1,10,12,  0,11,13,  12,14,19,  9,13,15,  4,14,16,  5,15,17,  6,16,18,  7,17,19,  8,13,18},

        //6-7 cycles with deg 2 edges.
        //{12,-1,-1,  7,-1,-1,  10,-1,-1,  8,-1,-1,  5,9,14,  4,8,-1,  7,14,-1,  1,6,8,  3,5,7,  4,10,-1,  2,9,11,  10,12,-1,  0,11,13,  12,14,-1,  4,6,13},
        //{12,-1,-1,  7,-1,-1,  9,-1,-1,  4,6,13,  3,5,9,  4,8,-1,  3,7,-1,  1,6,8,  5,7,-1,  2,4,10,  9,11,-1,  10,12,-1,  0,11,13,  3,12,-1},
        //{12,-1,-1,  7,-1,-1,  10,-1,-1,  4,6,13,  3,5,9,  4,8,-1,  3,7,-1,  1,6,8,  5,7,-1,  4,10,-1,  2,9,11,  10,12,-1,  0,11,13,  3,12,-1},
    };
    std::vector<std::vector<int>> mixedCycleTestOuts_one = {
        {20,9},
        //{15,4},{14,3},{14,3},
    };
    
    std::cout<<"Testing the 6-7 cycle with 1 common edge:" <<std::endl;
    groupTest(mixedCycleTest_one, mixedCycleTestOuts_one, true);

    std::vector<std::vector<int>> mixedCycleTest_two = {
        //Full 6-7 cycle with 2 common edges

        ////6-7 cycles with deg 2 edges.
        {12,-1,-1,  11,-1,-1,  10,-1,-1,  9,-1,-1,  15,-1,-1,  16,-1,-1,  17,-1,-1,  8,-1,-1,  7,12,17,  3,10,14,  2,9,11,  1,10,12,  0,8,11,13,  12,14,-1,  9,13,15,  4,14,16,  5,15,17,  6,8,16},
        {11,-1,-1,  10,-1,-1,  9,-1,-1,  8,-1,-1,  14,-1,-1,  15,-1,-1,  16,-1,-1,  11,16,-1,  3,9,13,  2,8,10,  1,9,11,  0,7,10,12,  11,13,-1,  8,12,14,  4,13,15,  5,14,16,  6,7,15},
        {10,-1,-1,  9,-1,-1,  8,-1,-1,  7,-1,-1,  13,-1,-1,  14,-1,-1,  10,15,-1,  3,8,12,  2,7,9,  1,8,10,  0,6,9,11,  10,12,-1,  7,11,13,  4,12,14,  5,13,15,  6,14,-1},
        {13,-1,-1,  11,-1,-1,  4,-1,-1,  6,-1,-1,  2,5,9,  4,6,-1,  3,5,7,  6,8,13,  7,9,-1,  4,8,10,  9,11,-1,  1,10,12,  11,13,-1,  0,7,12},
        {12,-1,-1,  7,-1,-1,  9,-1,-1,  4,6,12,  3,5,-1,  4,8,9,  3,7,-1,  1,6,8,  5,7,-1,  2,5,10,  9,11,-1,  10,12,-1,  0,3,11},
    };
    std::vector<std::vector<int>> mixedCycleTestOuts_two = {

        {18,8},{17, 7},{16,6},{14,4},{13,3},
    };

    // std::cout<<"Doing 6-7 cycle with 2 commons from 8 to 3 vertices:" <<std::endl;
    // groupTest(mixedCycleTest_two, mixedCycleTestOuts_two, false);

    std::vector<std::vector<int>> thetaCycleTest = {
        {5,-1,-1,  6,-1,-1,  7,-1,-1,  4,-1,-1,  3,9,12,  0,9,11,  1,8,11,  2,8,12,  6,7,-1,  4,5,-1,  11,12,-1,  5,6,10,  4,7,10},
        {3,-1,-1,  9,-1,-1,  11,-1,-1,  0,4,6,  3,5,7,  4,8,-1,  3,9,-1,  4,10,-1,  5,11,-1,  1,6,10,  7,9,11,  2,8,10},
    };
    std::vector<std::vector<int>> thetaCycleTestOuts = {
        {13,4},{12,3},
    };

    // std::cout<<"Doing Theta cycle with from 4 to 3 vertices:" <<std::endl;
    // groupTest(thetaCycleTest, thetaCycleTestOuts, false);
};

bool validityTests() {
    //Neighbor list for vertices indexed by increments of 3 (max degree).
    //Always starts with the outgoing vertices.
    std::vector<int> origEdge = {3,-1,-1,   4,-1,-1,  5,-1,-1,  0,4,6,  1,3,5,  2,4,6,   3,5,-1};
    std::vector<int> origOutVerts = {7,3}; 

    std::vector<int> badEdge = {3,-1,-1,   4,-1,-1,  5,-1,-1,  0,4,5,  1,3,5,  2,3,4};
    std::vector<int> badOutVerts = {6,3};

    std::vector<int> redEdge = {3,-1,-1,   3,-1,-1,  3,-1,-1,  0,1,2};
    std::vector<int> redOutVerts = {4,3};
    
    std::vector<int> outEdges = {2, -1, 0};

    Graph original(origOutVerts[0],origOutVerts[1], origEdge, outEdges);
    Graph reduced(redOutVerts[0],redOutVerts[1], redEdge, outEdges);
    Graph badling(6,3, badEdge, outEdges);

    std::vector<bool> origFactor = {1,0,0,  0,0,0,  1,0,0,  1,1,0,  0,1,1,  1,1,0,  0,0,0};
    std::vector<bool> redFactor = {1,0,0,  0,0,0,  1,0,0,  1,0,1};
    std::vector<bool> outFactor = {1,0,1};

    original.nextMask_reset();
    original.evalMetric(origFactor, outFactor);
    reduced.evalMetric(redFactor, outFactor);
    original.evalMetric(original.nextMask(outFactor), outFactor);
    original.evalMetric(original.nextMask(outFactor), outFactor);

    assert(original.evalMetric(origFactor, outFactor) == 7); 
    assert(reduced.evalMetric(redFactor, outFactor) == 7);    
    assert(original.nextMask(outFactor).empty());

//     Graph orig(origVertCount,3, origEdge, origOutVerts, {-1,2,1});
//     Graph red(redVertCount,3, redEdge, redOutVerts, {-1,2,1});
//     std::vector<bool> badFactor = {0,1,1};

//     orig.nextMask_reset();
//     red.nextMask_reset();

//     std::cout<<"Orig with a generated mask: " <<orig.evalMetric(orig.nextMask(badFactor), badFactor) <<std::endl
//              <<"Red with a generated mask: "  <<red.evalMetric(red.nextMask(badFactor), badFactor) <<std::endl;
    std::vector<int> quadEdge = {4,-1,-1,  5,-1,-1,  6,-1,-1,  7,-1,-1,  0,5,7,  1,4,6,  2,5,7,  3,4,6};
    std::vector<int> quadOutVerts = {8,4};
    std::vector<int> qOutEdges = {2, 3, 0, 1};
    std::vector<bool> qOutFactor = {1,1,1,1};

    Graph quad(quadOutVerts[0],quadOutVerts[1], quadEdge, qOutEdges);

    // quad.nextMask_reset();
    // std::vector<bool> foundFactor = quad.nextMask(qOutFactor);
    // while (!foundFactor.empty()) {
    //     for (auto i = foundFactor.begin(); i != foundFactor.end(); ++i)
    //         std::cout << *i << ' ';
    //     std::cout<<std::endl <<"The metric for this mask comes out at: " <<quad.evalMetric(foundFactor, qOutFactor) <<std::endl;
    //     foundFactor = quad.nextMask(qOutFactor);
    // }

    //===========================
    //Purely comparing graphs...
    //===========================
    std::vector<bool> factorTest;

    factorTest = Graph::findFactor(original, reduced);
    if (factorTest.empty()) return false;
    
    factorTest = Graph::findFactor(original, badling);
    if (!factorTest.empty()) return false;

    std::vector<int> rDiamondEdge = {2,-1,-1,  4,-1,-1,  0,3,4,  5,6,-1,  1,7,-1,  8,9,-1,  10,11,-1,  12,13,-1,  5,9,12,  5,8,12,  6,11,13,  6,10,13,  7,8,9,  7,10,11};

    std::vector<int> diamondEdge = {2,-1,-1,  3,-1,-1,  0,4,5,  1,4,5,  2,3,5,  2,3,4};
    std::vector<int> dOutEdges = {1,0};                  
    
    Graph diamond(6,2, diamondEdge, dOutEdges);
    Graph rDiamond(14,2, rDiamondEdge, dOutEdges);

    factorTest = Graph::findFactor(diamond, rDiamond);
    if (factorTest.empty()) return false;

    std::vector<int> triEdge = {2,-1,-1,  3,-1,-1,  0,3,4,  1,2,4,  2,3,-1};

    std::vector<int> rTriEdge = {2,-1,-1,  3,-1,-1,  0,3,-1,  1,2,-1,};
    std::vector<int> triEdges = {1,0};                  
    
    Graph triangle(5,2, triEdge, triEdges);
    Graph rTriangle(4,2, rTriEdge, triEdges);

    factorTest = Graph::findFactor(triangle, rTriangle);
    if (factorTest.empty()) return false;

    factorTest = Graph::findFactor(original, original);
    if (factorTest.empty()) return false;

    Graph largeT(forest::largeTestOuts[0],forest::largeTestOuts[1], forest::largeTest, {});

    // factorTest = Graph::findFactor(largeT, largeT);
    // runner = 0;
    // for (auto i = factorTest.begin(); i != factorTest.end(); ++i) {
    //         if (runner % 3 == 0)  std::cout << runner/3 << ": ";
    //         std::cout << *i << ' ';
    //         runner++;
    // }
    // std::cout <<std::endl;
    return true;
};

bool minorTests() {

    Graph sixer(12,6,
    {9,-1,-1,  10,-1,-1,  11,-1,-1,  6,-1,-1,  8,-1,-1,  7,-1,-1,  3,8,11,  5,8,9,  4,6,7,  0,7,10,  1,9,11,  2,6,10},
    {});

    Graph sameSixer(12,6,
    {10,-1,-1,  9,-1,-1,  11,-1,-1,  6,-1,-1,  7,-1,-1,  8,-1,-1,  3,8,11,  4,8,9,  5,6,7,  1,7,10,  0,9,11,  2,6,10},
    {});

    Graph triangleSixer(10,6,
    {7,-1,-1,9,-1,-1,8,-1,-1,6,-1,-1,8,-1,-1,6,-1,-1,3,5,7,2,6,-1,2,4,9,1,8,-1},
    {});

    Graph rSixer(7,6,
    {1,-1,-1,0,-1,-1,3,-1,-1,2,-1,-1,6,-1,-1,6,-1,-1,4,5,-1,},
    {});

    Graph duoSixer(11,5,
    {9,-1,-1,  10,-1,-1,  11,-1,-1,  6,-1,-1,  8,-1,-1,  3,8,11, 8,9,-1,  4,6,7,  0,7,10,  1,9,11,  2,6,10},
    {});

    Graph expandedSixer(14,7,
    {11,-1,-1,  10,-1,-1,  9,-1,-1,  8,-1,-1,  7,-1,-1,  13,-1,-1,  8,-1,-1,  4,9,13,  3,6,12,  2,7,10,  1,9,11,  0,10,12,  8,11,13,  5,7,12},
    {});

    Graph rExpandedSixer(9,7,
    {5,-1,-1,  7,-1,-1,  7,-1,-1,  8,-1,-1,  7,-1,-1,  0,-1,-1,  8,-1,-1,  1,2,4,  3,6,-1},
    {-1,5,4,-1,2,1,-1});

    Graph nonEdgedSixer(6,6,
    {1,-1,-1,  0,-1,-1,  3,-1,-1,  2,-1,-1,  5,-1,-1,  4,-1,-1},
    {});

    Graph isolateSixer(6,6,
    {1,-1,-1,  0,-1,-1,  2,-1,-1,  3,-1,-1,  4,-1,-1,  5,-1,-1},
    {});

    Graph edgedSixer(9,6,
    {7,-1,-1,  7,-1,-1,  6,-1,-1,  6,-1,-1,  8,-1,-1,  8,-1,-1,  2,3,-1,  0,1,-1,  4,5,-1},
    {});


    // std::cout<<"Testing the 6-6 cycle:" <<std::endl;
    // do {
    //     treeTest(sixer);
    // } while (sixer.permuteEdges());

    // rExpandedSixer.nextMask_reset();
    // std::vector<bool> mask = rExpandedSixer.nextMask({0,1,1,0,1,1,0});

    if(Graph::findFactor(sixer, triangleSixer).empty()) {
        return false;
    }

    if(!Graph::findFactor(sixer, rSixer).empty() || 
       !Graph::findFactor(sameSixer, rSixer).empty()) {
        return false;
    }

    if(Graph::findFactor(expandedSixer, rExpandedSixer).empty()) {
        return false;
    }

    //Disabled unless we add isolates!
    // if(Graph::findFactor(sixer, isolateSixer).empty()) 
    //     return false;
    
    Graph Sevener(14,7,
    {7,-1,-1,13,-1,-1,8,-1,-1,9,-1,-1,10,-1,-1,11,-1,-1,12,-1,-1,0,8,13,2,7,9,3,8,10,4,9,11,5,10,12,6,11,13,1,7,12,},
    {});
    
    Graph reducedSevener(8,7,
    {1,-1,-1,0,-1,-1,3,-1,-1,2,-1,-1,7,-1,-1,7,-1,-1,7,-1,-1,4,5,6,},
    {});
    
    // std::cout<<"Testing the 7-7 cycle:" <<std::endl;
    // if(!Graph::findFactor(Sevener, reducedSevener).empty()) 
    //     return false;

    Graph Giant(20,10,
    {19,-1,-1,  10,-1,-1,  11,-1,-1,  12,-1,-1,  13,-1,-1,  14,-1,-1,  15,-1,-1,  16,-1,-1,  17,-1,-1,  18,-1,-1,  1,11,19,  2,10,12,  3,11,13,  4,12,14,  5,13,15,  6,14,16,  7,15,17,  8,16,18,  9,17,19,  0,10,18},
    {});

    Graph reducedGiant(10,10,
    {1,-1,-1,  0,-1,-1,  3,-1,-1,  2,-1,-1,  5,-1,-1,  4,-1,-1, 7,-1,-1,  6,-1,-1, 9,-1,-1,  8,-1,-1,},
    {});

    // std::cout<<"Testing the reduced 10 cycle:" <<std::endl;
    // if(!Graph::findFactor(Giant, reducedGiant).empty()) 
    //     return false;

    return true;
}